package com.bergerkiller.bukkit.nolagg;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.bukkit.Location;
import org.bukkit.entity.Entity;
import org.bukkit.entity.ExperienceOrb;
import org.bukkit.entity.Item;
import org.bukkit.inventory.ItemStack;
import org.bukkit.util.Vector;

import com.bergerkiller.bukkit.nolagg.meta.WorldMetaData;

public class StackFormer extends Thread {
	public static double stackRadiusSquared;
	public static int stackThreshold = 2;
	private static StackFormer thread;
	
	private static Set<Entity> entitiesToKill = new HashSet<Entity>();
	private static Set<Item> itemsToRespawn = new HashSet<Item>();
	private static final List<Item> items = new ArrayList<Item>();
	private static final List<ExperienceOrb> orbs = new ArrayList<ExperienceOrb>();
	
	private static boolean orbrunning = false;
	private static boolean itemrunning = false;
	public static void update(List<Item> itemlist, List<ExperienceOrb> orblist) {
		if (!ItemHandler.formStacks) return;
		if (stackThreshold < 2) return;
		synchronized (thread) {
			if (itemrunning || orbrunning) return;
			for (Entity e : entitiesToKill)  {
			    if (e instanceof Item) WorldMetaData.removeSpawnedItem((Item) e);
				e.remove();
			}
			entitiesToKill.clear();
			for (Item item : itemsToRespawn) {
				ItemHandler.respawnItem(item, new Vector());
			}
			itemsToRespawn.clear();
			if (!orbrunning) {
				orbs.addAll(orblist);
				orbrunning = true;
			}
			if (!itemrunning) {
				items.addAll(itemlist);
				itemrunning = true;
			}
		}
	}
		
	private static void updateOrbs() {
		for (ExperienceOrb orb : orbs) {
			if (orb.isDead()) continue;
			if (addSameOrbsNear(orbs, orb)) {
				if (near.size() > stackThreshold - 2) {
					for (Entity e : near) {
						if (e.isDead()) continue;
						//add the experience
						ExperienceOrb to = (ExperienceOrb) e;
						orb.setExperience(orb.getExperience() + to.getExperience());
						entitiesToKill.add(e);
					}						
				}
				near.clear();
			}
		}
		orbs.clear();
	}
	private static void updateItems() {
		ItemStack stack;
		int maxsize;
		for (Item item : items) {
			if (entitiesToKill.contains(item)) continue;
			stack = item.getItemStack();
			maxsize = stack.getMaxStackSize();
			if (stack.getAmount() >= maxsize) continue;
			if (addSameItemsNear(items, item)) {
				if (near.size() > stackThreshold - 2) {
					//addition the items
					Item nearitem;
					ItemStack nearitemstack;
					boolean changed = false;
					for (Entity e : near) {
						if (entitiesToKill.contains(e)) continue;
						nearitem = (Item) e;
						nearitemstack = nearitem.getItemStack();
						if (nearitemstack == null) continue;
						if (nearitemstack.getAmount() >= maxsize) continue;
						int newamount = stack.getAmount() + nearitemstack.getAmount();
						if (newamount <= maxsize) {
							if (newamount == 0) newamount = 1;
							stack.setAmount(newamount);
							entitiesToKill.add(nearitem);
							changed = true;
						} else {
							//set to max
							stack.setAmount(maxsize);
							//set prev. item
							nearitemstack.setAmount(newamount - maxsize);
							changed = true;
						}
						
					}
					if (changed) {
						itemsToRespawn.add(item);
					}
				}
				near.clear();
			}
		}
		items.clear();
	}
	
	private static List<Entity> near = new ArrayList<Entity>(stackThreshold - 1);
	private static boolean addSameItemsNear(List<Item> from, Item around) {
		if (around.isDead()) return false;
		Location m = around.getLocation();
		ItemStack stack = around.getItemStack();
		boolean added = false;
		for (Item item : from) {
			if (item.isDead()) continue;
			if (item == around) continue;
			if (canStack(m, item.getLocation())) {
				ItemStack instack = item.getItemStack();
				if (instack.getTypeId() != stack.getTypeId()) {
					continue;
				}
				if (instack.getDurability() != stack.getDurability()) {
					continue;
				}
				added = true;
				near.add(item);
			}
		}
		return added;
	}
	private static boolean addSameOrbsNear(List<ExperienceOrb> from, ExperienceOrb around) {
		if (around.isDead()) return false;
		Location m = around.getLocation();
		boolean added = true;
		for (ExperienceOrb orb : from) {
			if (orb.isDead()) continue;
			if (orb == around) continue;
			if (canStack(m, orb.getLocation())) {
				added = true;
				near.add(orb);
			}
		}
		return added;
	}
	private static boolean canStack(Location l1, Location l2) {
		double d = distance(l1.getX(), l2.getX());
		if (d > stackRadiusSquared) return false;
		d += distance(l1.getZ(), l2.getZ());
		if (d > stackRadiusSquared) return false;
		d += distance(l1.getY(), l2.getY());
		if (d > stackRadiusSquared) return false;
		return true;
	}	
	private static double distance(double d1, double d2) {
		d1 = Math.abs(d1 - d2);
		return d1 * d1;
	}

	public static void init() {
		thread = new StackFormer();
		thread.start();
	}
	public static void deinit() {
		thread.running = false;
		thread = null;
	}
	boolean running = true;
	public void run() {
		boolean done;
		while (!this.isInterrupted() && this.running) {
			try {
				synchronized (thread) {
					done = itemrunning;
				}
				if (done) {
					updateItems();
					synchronized (thread) {
						itemrunning = false;
					}
				}
				
				synchronized (thread) {
					done = orbrunning;
				}
				if (done) {
					updateOrbs();
					synchronized (thread) {
						orbrunning = false;
					}
				}
				Thread.sleep(200);
			} catch (InterruptedException ex) {
			} catch (Exception ex) {
				if (this.running) {
					//log
					ex.printStackTrace();
				}
			}
		}
	}
	

	
}
